'''
Neuron simulator export for:

Components:
    null (Type: notes)
    DoubExpSyn (Type: expTwoSynapse:  tauRise=0.001 (SI time) tauDecay=0.002 (SI time) peakTime=0.0013862943611198907 (SI time) waveformFactor=4.0 (dimensionless) gbase=1.0E-8 (SI conductance) erev=0.0 (SI voltage))
    null (Type: notes)
    KConductance (Type: ionChannelHH:  conductance=1.0E-11 (SI conductance))
    null (Type: notes)
    LeakConductance (Type: ionChannelPassive:  conductance=1.0E-11 (SI conductance))
    null (Type: notes)
    NaConductance (Type: ionChannelHH:  conductance=1.0E-11 (SI conductance))
    TestCell_ChannelML (Type: cell)
    null (Type: notes)
    Input_0 (Type: pulseGenerator:  delay=0.02 (SI time) duration=0.06 (SI time) amplitude=2.0E-10 (SI current))
    network_Ex4_HHcell (Type: networkWithTemperature:  temperature=279.45 (SI temperature))
    sim1 (Type: Simulation:  length=0.1 (SI time) step=5.0E-6 (SI time))


    This NEURON file has been generated by org.neuroml.export (see https://github.com/NeuroML/org.neuroml.export)
         org.neuroml.export  v1.5.5
         org.neuroml.model   v1.5.5
         jLEMS               v0.9.9.2

'''

import neuron

import time
import sys

import hashlib
h = neuron.h
h.load_file("stdlib.hoc")

h.load_file("stdgui.hoc")

h("objref p")
h("p = new PythonObject()")

class NeuronSimulation():

    def __init__(self, tstop, dt, seed=123456789):

        print("\n    Starting simulation in NEURON of %sms generated from NeuroML2 model...\n"%tstop)

        setup_start = time.time()
        self.report_file = open('simulator.props','w')
        print('Simulator version:  %s'%h.nrnversion())
        self.report_file.write('# Report of running simulation with %s\n'%h.nrnversion())
        self.report_file.write('SimulationFile=%s\n'%__file__)
        self.report_file.write('PythonVersion=%s\n'%sys.version.replace('\n',' '))
        print('Python version:     %s'%sys.version.replace('\n',' '))
        self.report_file.write('SimulatorVersion=%s\n'%h.nrnversion())

        self.seed = seed
        self.randoms = []
        self.next_global_id = 0  # Used in Random123 classes for elements using random(), etc. 

        self.next_spiking_input_id = 0  # Used in Random123 classes for elements using random(), etc. 

        '''
        Adding simulation Component(id=sim1 type=Simulation) of network/component: network_Ex4_HHcell (Type: networkWithTemperature:  temperature=279.45 (SI temperature))
        
        '''

        # Temperature used for network: 279.45 K
        h.celsius = 279.45 - 273.15

        # ######################   Population: CG1
        print("Population CG1 contains 1 instance(s) of component: TestCell_ChannelML of type: cell")

        h.load_file("TestCell_ChannelML.hoc")
        a_CG1 = []
        h("{ n_CG1 = 1 }")
        h("objectvar a_CG1[n_CG1]")
        for i in range(int(h.n_CG1)):
            h("a_CG1[%i] = new TestCell_ChannelML()"%i)
            h("access a_CG1[%i].Soma"%i)

            self.next_global_id+=1

        h("{ a_CG1[0].position(54.987995, 27.486061, 40.31429) }")

        h("proc initialiseV_CG1() { for i = 0, n_CG1-1 { a_CG1[i].set_initial_v() } }")
        h("objref fih_CG1")
        h('{fih_CG1 = new FInitializeHandler(0, "initialiseV_CG1()")}')

        h("proc initialiseIons_CG1() { for i = 0, n_CG1-1 { a_CG1[i].set_initial_ion_properties() } }")
        h("objref fih_ion_CG1")
        h('{fih_ion_CG1 = new FInitializeHandler(1, "initialiseIons_CG1()")}')

        # ######################   Population: CG2
        print("Population CG2 contains 1 instance(s) of component: TestCell_ChannelML of type: cell")

        h.load_file("TestCell_ChannelML.hoc")
        a_CG2 = []
        h("{ n_CG2 = 1 }")
        h("objectvar a_CG2[n_CG2]")
        for i in range(int(h.n_CG2)):
            h("a_CG2[%i] = new TestCell_ChannelML()"%i)
            h("access a_CG2[%i].Soma"%i)

            self.next_global_id+=1

        h("{ a_CG2[0].position(41.628883, 72.79404, 74.07768) }")

        h("proc initialiseV_CG2() { for i = 0, n_CG2-1 { a_CG2[i].set_initial_v() } }")
        h("objref fih_CG2")
        h('{fih_CG2 = new FInitializeHandler(0, "initialiseV_CG2()")}')

        h("proc initialiseIons_CG2() { for i = 0, n_CG2-1 { a_CG2[i].set_initial_ion_properties() } }")
        h("objref fih_ion_CG2")
        h('{fih_ion_CG2 = new FInitializeHandler(1, "initialiseIons_CG2()")}')

        # ######################   Projection: NetConn_1
        print("Adding projection: NetConn_1, from CG1 to CG2 with synapse DoubExpSyn, 1 connection(s)")

        h("objectvar syn_NetConn_1_DoubExpSyn[1]")

        h("objectvar netConn_NetConn_1_DoubExpSyn[1]")

        # Connection 0: cell 0, seg 0 (0.5) [0.5 on a_CG1[0].Soma] -> cell 0, seg 0 (0.5) [0.5 on a_CG2[0].Soma], weight: 1.0, delay 0.0
        h("a_CG2[0].Soma syn_NetConn_1_DoubExpSyn[0] = new DoubExpSyn(0.5)")
        h("a_CG1[0].Soma a_CG2[0].synlist.append(new NetCon(&v(0.5), syn_NetConn_1_DoubExpSyn[0], -20.0, 0.0, 1.0))")

        print("Processing 1 input lists")

        # ######################   Input List: Input_0
        # Adding single input: Component(id=0 type=input)
        h("objref Input_0_0")
        h("a_CG1[0].Soma { Input_0_0 = new Input_0(0.5) } ")

        print("Finished processing 1 input lists")

        trec = h.Vector()
        trec.record(h._ref_t)

        h.tstop = tstop

        h.dt = dt

        h.steps_per_ms = 1/h.dt



        # ######################   File to save: CG1_0.0.KConductance_COND_DENS_k.dat (Var_1_0_0_OF)
        # Column: CG1/0/TestCell_ChannelML/biophys/membraneProperties/KConductance_all/gDensity
        h(' objectvar v_gDensity_Var_1_0_0_OF ')
        h(' { v_gDensity_Var_1_0_0_OF = new Vector() } ')
        h(' { v_gDensity_Var_1_0_0_OF.record(&a_CG1[0].Soma.gion_KConductance(0.5)) } ')
        h.v_gDensity_Var_1_0_0_OF.resize((h.tstop * h.steps_per_ms) + 1)

        # ######################   File to save: time.dat (time)
        # Column: time
        h(' objectvar v_time ')
        h(' { v_time = new Vector() } ')
        h(' { v_time.record(&t) } ')
        h.v_time.resize((h.tstop * h.steps_per_ms) + 1)

        # ######################   File to save: CG2_0.0.dat (Plot_1_0_0_OF)
        # Column: CG2/0/TestCell_ChannelML/v
        h(' objectvar v_v_Plot_1_0_0_OF ')
        h(' { v_v_Plot_1_0_0_OF = new Vector() } ')
        h(' { v_v_Plot_1_0_0_OF.record(&a_CG2[0].Soma.v(0.5)) } ')
        h.v_v_Plot_1_0_0_OF.resize((h.tstop * h.steps_per_ms) + 1)

        # ######################   File to save: CG1_0.0.NaConductance_COND_DENS_na.dat (Var_0_0_0_OF)
        # Column: CG1/0/TestCell_ChannelML/biophys/membraneProperties/NaConductance_all/gDensity
        h(' objectvar v_gDensity_Var_0_0_0_OF ')
        h(' { v_gDensity_Var_0_0_0_OF = new Vector() } ')
        h(' { v_gDensity_Var_0_0_0_OF.record(&a_CG1[0].Soma.gion_NaConductance(0.5)) } ')
        h.v_gDensity_Var_0_0_0_OF.resize((h.tstop * h.steps_per_ms) + 1)

        # ######################   File to save: CG1_0.0.dat (Plot_0_0_0_OF)
        # Column: CG1/0/TestCell_ChannelML/v
        h(' objectvar v_v_Plot_0_0_0_OF ')
        h(' { v_v_Plot_0_0_0_OF = new Vector() } ')
        h(' { v_v_Plot_0_0_0_OF.record(&a_CG1[0].Soma.v(0.5)) } ')
        h.v_v_Plot_0_0_0_OF.resize((h.tstop * h.steps_per_ms) + 1)

        # ######################   File to save: CG1_0.0.KConductance_n.dat (Plot_2_0_0_OF)
        # Column: CG1/0/TestCell_ChannelML/biophys/membraneProperties/KConductance_all/KConductance/n/q
        h(' objectvar v_q_Plot_2_0_0_OF ')
        h(' { v_q_Plot_2_0_0_OF = new Vector() } ')
        h(' { v_q_Plot_2_0_0_OF.record(&a_CG1[0].Soma.n_q_KConductance(0.5)) } ')
        h.v_q_Plot_2_0_0_OF.resize((h.tstop * h.steps_per_ms) + 1)

        # ######################   File to save: CG1_0.0.NaConductance_m.dat (Plot_3_0_0_OF)
        # Column: CG1/0/TestCell_ChannelML/biophys/membraneProperties/NaConductance_all/NaConductance/m/q
        h(' objectvar v_q_Plot_3_0_0_OF ')
        h(' { v_q_Plot_3_0_0_OF = new Vector() } ')
        h(' { v_q_Plot_3_0_0_OF.record(&a_CG1[0].Soma.m_q_NaConductance(0.5)) } ')
        h.v_q_Plot_3_0_0_OF.resize((h.tstop * h.steps_per_ms) + 1)

        # ######################   File to save: CG1_0.0.NaConductance_h.dat (Plot_4_0_0_OF)
        # Column: CG1/0/TestCell_ChannelML/biophys/membraneProperties/NaConductance_all/NaConductance/h/q
        h(' objectvar v_q_Plot_4_0_0_OF ')
        h(' { v_q_Plot_4_0_0_OF = new Vector() } ')
        h(' { v_q_Plot_4_0_0_OF.record(&a_CG1[0].Soma.h_q_NaConductance(0.5)) } ')
        h.v_q_Plot_4_0_0_OF.resize((h.tstop * h.steps_per_ms) + 1)

        self.initialized = False

        self.sim_end = -1 # will be overwritten

        setup_end = time.time()
        self.setup_time = setup_end - setup_start
        print("Set up network to simulate in %f seconds"%(self.setup_time))

    def run(self):

        self.initialized = True
        sim_start = time.time()
        print("Running a simulation of %sms (dt = %sms; seed=%s)" % (h.tstop, h.dt, self.seed))

        try:
            h.run()
        except Exception as e:
            print("Exception running NEURON: %s" % (e))
            quit()


        self.sim_end = time.time()
        self.sim_time = self.sim_end - sim_start
        print("Finished NEURON simulation in %f seconds (%f mins)..."%(self.sim_time, self.sim_time/60.0))

        try:
            self.save_results()
        except Exception as e:
            print("Exception saving results of NEURON simulation: %s" % (e))
            quit()


    def advance(self):

        if not self.initialized:
            h.finitialize()
            self.initialized = True

        h.fadvance()


    ###############################################################################
    # Hash function to use in generation of random value
    # This is copied from NetPyNE: https://github.com/Neurosim-lab/netpyne/blob/master/netpyne/simFuncs.py
    ###############################################################################
    def _id32 (self,obj): 
        return int(hashlib.md5(obj).hexdigest()[0:8],16)  # convert 8 first chars of md5 hash in base 16 to int


    ###############################################################################
    # Initialize the stim randomizer
    # This is copied from NetPyNE: https://github.com/Neurosim-lab/netpyne/blob/master/netpyne/simFuncs.py
    ###############################################################################
    def _init_stim_randomizer(self,rand, stimType, gid, seed): 
        #print("INIT STIM  %s; %s; %s; %s"%(rand, stimType, gid, seed))
        rand.Random123(self._id32(stimType), gid, seed)


    def save_results(self):

        print("Saving results at t=%s..."%h.t)

        if self.sim_end < 0: self.sim_end = time.time()


        # ######################   File to save: time.dat (time)
        py_v_time = [ t/1000 for t in h.v_time.to_python() ]  # Convert to Python list for speed...

        f_time_f2 = open('time.dat', 'w')
        num_points = len(py_v_time)  # Simulation may have been stopped before tstop...

        for i in range(num_points):
            f_time_f2.write('%f'% py_v_time[i])  # Save in SI units...
        f_time_f2.close()
        print("Saved data to: time.dat")

        # ######################   File to save: CG1_0.0.KConductance_COND_DENS_k.dat (Var_1_0_0_OF)
        py_v_gDensity_Var_1_0_0_OF = [ float(x  / 1.0E-6) for x in h.v_gDensity_Var_1_0_0_OF.to_python() ]  # Convert to Python list for speed, variable has dim: conductanceDensity

        f_Var_1_0_0_OF_f2 = open('CG1_0.0.KConductance_COND_DENS_k.dat', 'w')
        num_points = len(py_v_time)  # Simulation may have been stopped before tstop...

        for i in range(num_points):
            f_Var_1_0_0_OF_f2.write('%e\t%e\t\n' % (py_v_time[i], py_v_gDensity_Var_1_0_0_OF[i], ))
        f_Var_1_0_0_OF_f2.close()
        print("Saved data to: CG1_0.0.KConductance_COND_DENS_k.dat")

        # ######################   File to save: CG2_0.0.dat (Plot_1_0_0_OF)
        py_v_v_Plot_1_0_0_OF = [ float(x  / 1000.0) for x in h.v_v_Plot_1_0_0_OF.to_python() ]  # Convert to Python list for speed, variable has dim: voltage

        f_Plot_1_0_0_OF_f2 = open('CG2_0.0.dat', 'w')
        num_points = len(py_v_time)  # Simulation may have been stopped before tstop...

        for i in range(num_points):
            f_Plot_1_0_0_OF_f2.write('%e\t%e\t\n' % (py_v_time[i], py_v_v_Plot_1_0_0_OF[i], ))
        f_Plot_1_0_0_OF_f2.close()
        print("Saved data to: CG2_0.0.dat")

        # ######################   File to save: CG1_0.0.NaConductance_COND_DENS_na.dat (Var_0_0_0_OF)
        py_v_gDensity_Var_0_0_0_OF = [ float(x  / 1.0E-6) for x in h.v_gDensity_Var_0_0_0_OF.to_python() ]  # Convert to Python list for speed, variable has dim: conductanceDensity

        f_Var_0_0_0_OF_f2 = open('CG1_0.0.NaConductance_COND_DENS_na.dat', 'w')
        num_points = len(py_v_time)  # Simulation may have been stopped before tstop...

        for i in range(num_points):
            f_Var_0_0_0_OF_f2.write('%e\t%e\t\n' % (py_v_time[i], py_v_gDensity_Var_0_0_0_OF[i], ))
        f_Var_0_0_0_OF_f2.close()
        print("Saved data to: CG1_0.0.NaConductance_COND_DENS_na.dat")

        # ######################   File to save: CG1_0.0.dat (Plot_0_0_0_OF)
        py_v_v_Plot_0_0_0_OF = [ float(x  / 1000.0) for x in h.v_v_Plot_0_0_0_OF.to_python() ]  # Convert to Python list for speed, variable has dim: voltage

        f_Plot_0_0_0_OF_f2 = open('CG1_0.0.dat', 'w')
        num_points = len(py_v_time)  # Simulation may have been stopped before tstop...

        for i in range(num_points):
            f_Plot_0_0_0_OF_f2.write('%e\t%e\t\n' % (py_v_time[i], py_v_v_Plot_0_0_0_OF[i], ))
        f_Plot_0_0_0_OF_f2.close()
        print("Saved data to: CG1_0.0.dat")

        # ######################   File to save: CG1_0.0.KConductance_n.dat (Plot_2_0_0_OF)
        py_v_q_Plot_2_0_0_OF = [ float(x ) for x in h.v_q_Plot_2_0_0_OF.to_python() ]  # Convert to Python list for speed, variable has dim: none

        f_Plot_2_0_0_OF_f2 = open('CG1_0.0.KConductance_n.dat', 'w')
        num_points = len(py_v_time)  # Simulation may have been stopped before tstop...

        for i in range(num_points):
            f_Plot_2_0_0_OF_f2.write('%e\t%e\t\n' % (py_v_time[i], py_v_q_Plot_2_0_0_OF[i], ))
        f_Plot_2_0_0_OF_f2.close()
        print("Saved data to: CG1_0.0.KConductance_n.dat")

        # ######################   File to save: CG1_0.0.NaConductance_m.dat (Plot_3_0_0_OF)
        py_v_q_Plot_3_0_0_OF = [ float(x ) for x in h.v_q_Plot_3_0_0_OF.to_python() ]  # Convert to Python list for speed, variable has dim: none

        f_Plot_3_0_0_OF_f2 = open('CG1_0.0.NaConductance_m.dat', 'w')
        num_points = len(py_v_time)  # Simulation may have been stopped before tstop...

        for i in range(num_points):
            f_Plot_3_0_0_OF_f2.write('%e\t%e\t\n' % (py_v_time[i], py_v_q_Plot_3_0_0_OF[i], ))
        f_Plot_3_0_0_OF_f2.close()
        print("Saved data to: CG1_0.0.NaConductance_m.dat")

        # ######################   File to save: CG1_0.0.NaConductance_h.dat (Plot_4_0_0_OF)
        py_v_q_Plot_4_0_0_OF = [ float(x ) for x in h.v_q_Plot_4_0_0_OF.to_python() ]  # Convert to Python list for speed, variable has dim: none

        f_Plot_4_0_0_OF_f2 = open('CG1_0.0.NaConductance_h.dat', 'w')
        num_points = len(py_v_time)  # Simulation may have been stopped before tstop...

        for i in range(num_points):
            f_Plot_4_0_0_OF_f2.write('%e\t%e\t\n' % (py_v_time[i], py_v_q_Plot_4_0_0_OF[i], ))
        f_Plot_4_0_0_OF_f2.close()
        print("Saved data to: CG1_0.0.NaConductance_h.dat")

        save_end = time.time()
        save_time = save_end - self.sim_end
        print("Finished saving results in %f seconds"%(save_time))

        self.report_file.write('SetupTime=%s\n'%self.setup_time)
        self.report_file.write('RealSimulationTime=%s\n'%self.sim_time)
        self.report_file.write('SimulationSaveTime=%s\n'%save_time)
        self.report_file.close()

        print("Saving report of simulation to %s"%('simulator.props'))

        print("Done")

        quit()


if __name__ == '__main__':

    ns = NeuronSimulation(tstop=100.0, dt=0.005, seed=123456789)

    ns.run()

